/*
Copyright (c) 2016 David Gow <david@ingeniumdigital.com> 

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE. 
*/
#include "sprite.h"

#include <SDL.h>
#include <GL/glew.h>
#include <GL/glu.h>

#include "map.h"
#include <vector>

#define PLAYERSIZE 0.15

Sprite player = {2,2,0.15,3.14159265/2.0,1.0,1.0,1.0};

std::vector<Sprite*> g_Sprites;

GLuint shape_prog;
GLuint spuparam, healthparam;

void PLAYER_DrawOverlay()
{
	glDisable(GL_DEPTH_TEST);
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0,1366, 768, 0);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glUseProgram(shape_prog);
	glUniform1f(spuparam, player.type);
	glUniform1f(healthparam, player.health);
	glBegin(GL_QUADS);
	glTexCoord2f(0,0);
	glVertex3f(1300, 700, 0);
	glTexCoord2f(0,1);
	glVertex3f(1300, 766,0);
	glTexCoord2f(1,1);
	glVertex3f(1366, 766,0);
	glTexCoord2f(1,0);
	glVertex3f(1366, 700, 0);
	glEnd();
	glUseProgram(0);
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}

void SPRITE_LoadAssets()
{
	const char *shapeShaderSource =
		"#version 120\n"
		"\n"
		"uniform float camt;\n"
		"uniform float health;\n"
		"\n"
		"void main() {\n"
		"\t//Triangle\n"
		"\tfloat ds = abs(gl_TexCoord[0].x - 0.5);\n"
		"\tfloat sdf = -ds * 2 + gl_TexCoord[0].y;\n"
		"\t//Circle\n"
		"\tvec2 cs = vec2(gl_TexCoord[0].x - 0.5, gl_TexCoord[0].y - 0.5);\n"
		"\tfloat csdf = (0.5 - length(cs));\n"
		"\tfloat rsdf = mix(csdf, sdf, camt);\n"
		"\tif (rsdf < 0) discard;\n"
		"\tgl_FragColor = vec4(1,health,health,1);\n"
		"}\n";
	GLuint shape_shader = glCreateShader(GL_FRAGMENT_SHADER);
	glShaderSource(shape_shader, 1, &shapeShaderSource, 0);
	glCompileShader(shape_shader);
	shape_prog = glCreateProgram();
	glAttachShader(shape_prog, shape_shader);
	glLinkProgram(shape_prog);
	spuparam = glGetUniformLocation(shape_prog,"camt");
	healthparam = glGetUniformLocation(shape_prog,"health");
}

void SPRITE_Draw(Sprite *spr)
{
	float x = spr->x;
	float y = spr->y;
	float player_angle = player.angle;
	glUseProgram(shape_prog);
	glUniform1f(spuparam, spr->type);
	glUniform1f(healthparam, spr->health);
	glBegin(GL_QUADS);
	glTexCoord2f(0,1);
	glVertex3f(x-spr->size*sin(-player_angle), -spr->size, y-spr->size*cos(-player_angle));
	glTexCoord2f(1,1);
	glVertex3f(x+spr->size*sin(-player_angle), -spr->size, y+spr->size*cos(-player_angle));
	glTexCoord2f(1,0);
	glVertex3f(x+spr->size*sin(-player_angle), spr->size, y+spr->size*cos(-player_angle));
	glTexCoord2f(0,0);
	glVertex3f(x-spr->size*sin(-player_angle), spr->size, y-spr->size*cos(-player_angle));
	glEnd();
	glUseProgram(0);
}

// Does an object of size PLAYERSIZE at (px, py) collide with a tile at (x,y)
bool ClipPoint(int x, int y, float px, float py)
{
	float fx = float(x), fy = float(y);
	if (px < fx - 0.5 - PLAYERSIZE) return false; 
	if (px > fx + 0.5 + PLAYERSIZE) return false; 
	if (py < fy - 0.5 - PLAYERSIZE) return false; 
	if (py > fy + 0.5 + PLAYERSIZE) return false; 
	return true;
}

bool CanMove(float px, float py)
{
	for (int y = 0; y < 22; ++y)
	{
		for (int x = 0; x < 16; ++x)
		{
			if (MAP_IsSolid(x,y))
			{
				if (ClipPoint(x,y, px, py))
					return false;
			}
		}
	}
	return true;
}

void SPRITE_Move(Sprite *spr, float dx, float dy)
{
	if (CanMove(spr->x + dx, spr->y+dy))
	{
		spr->x += dx;
		spr->y += dy;
	}
	else if (CanMove(spr->x + dx, spr->y))
	{
		spr->x += dx;
	}
	else if (CanMove(spr->x, spr->y +dy))
	{
		spr->y += dy;
	}
}

void SPRITE_Thrust(Sprite *spr, float amt, float angle)
{
	SPRITE_Move(spr, cos(angle) * amt, sin(angle) * amt);
}

bool SPRITE_IsEnemy(Sprite *a, Sprite *b)
{
	return (a->type < 0.5) == (b->type > 0.5);
}

float EuclidDistSqr(float x1, float y1, float x2, float y2)
{
	return (x1 - x2)*(x1 - x2) + (y1 - y2)*(y1 - y2);
}

float EuclidDist(float x1, float y1, float x2, float y2)
{
	return SDL_sqrtf(EuclidDistSqr(x1,y1,x2,y2));
}

void Normalize(float& x, float& y)
{
	float len = EuclidDist(x,y,0,0);
	x /= len;
	y /= len;
}

void Sprite::CollideWith(Sprite* other)
{
	if (SPRITE_IsEnemy(this, other))
		other->health -= 0.01;
}

void SPRITE_DoAllCollisions()
{
	for (int i = 0; i < g_Sprites.size(); ++i)
	{
		for (int j = i+1; j < g_Sprites.size(); ++j)
		{
			Sprite *a = g_Sprites[i];
			Sprite *b = g_Sprites[j];
			float square_dist = EuclidDistSqr(a->x, a->y, b->x,b->y);
			float desired_dist = (a->size + b->size);
			if (square_dist < desired_dist*desired_dist)
			{
				// a and b are colliding.
				a->CollideWith(b);
				b->CollideWith(a);
				float a_to_b_x = b->x - a->x;
				float a_to_b_y = b->y - a->y;
				float current_dist = sqrtf(square_dist);
				float correction_dist = desired_dist - current_dist;
				printf("Collision: (%f, %f) dist %f - %f = %f\n", a_to_b_x, a_to_b_y, desired_dist, current_dist, correction_dist);
				Normalize(a_to_b_x, a_to_b_y);
				// pushable
				if (a != &player && b != &player)
				{
					SPRITE_Move(a, a_to_b_x * -correction_dist * 0.5, a_to_b_y * -correction_dist * 0.5);
					SPRITE_Move(b, a_to_b_x * correction_dist * 0.5, a_to_b_y * correction_dist * 0.5);
				}
				else
				{
					SPRITE_Move(a, a_to_b_x * -correction_dist, a_to_b_y * -correction_dist);
				}
			}
		}
	}
}

void PLAYER_Update()
{
	Sprite *spr = &player;
	if (fabs(spr->type - spr->desired_type) > 0.001)
	{
		spr->type = 0.9*(spr->type) + 0.1*(spr->desired_type);
	}
	printf("type = %f, desired = %f\n", spr->type, spr->desired_type);
}

void SPRITE_Update(Sprite *spr)
{
	if (spr->health <= 0.0) return;
	if (fabs(spr->type - spr->desired_type) > 0.1)
	{
		spr->type = 0.9*(spr->type) + 0.1*(spr->desired_type);
	}
	float minDist = EuclidDist(spr->x, spr->y, player.x, player.y);
	Sprite *closestE = &player;
	for (int i = 0; i < g_Sprites.size(); ++i)
	{
		Sprite *other = g_Sprites[i];
		if (MAP_LineIntersectsSolid(closestE->x, closestE->y, spr->x, spr->y))
			continue;
		if (SPRITE_IsEnemy(spr, other))
		{
			if (EuclidDist(spr->x, spr->y, other->x, other->y) < minDist)
			{
				minDist = EuclidDist(spr->x, spr->y, other->x, other->y);
				closestE = other;
			}
		}
	}
	if (!SPRITE_IsEnemy(spr, closestE))
		return;
	float mx = closestE->x - spr->x;
	float my = closestE->y - spr->y;
	// Only chase enemies we can see.
	if (MAP_LineIntersectsSolid(closestE->x, closestE->y, spr->x, spr->y))
		return;
	Normalize(mx, my);
	mx *= 0.01; my *= 0.01;
	SPRITE_Move(spr, mx, my);
}

void SPRITE_ToggleType(Sprite *spr)
{
	spr->desired_type = 1.0 - spr->desired_type;
}

void SPRITE_DrawAll()
{
	for (int i = 0; i < g_Sprites.size(); ++i)
	{
		SPRITE_Update(g_Sprites[i]);
		SPRITE_Draw(g_Sprites[i]);
	}
}

void SPRITE_InitMap()
{
	for (int i = 1; i < g_Sprites.size(); ++i)
	{
		delete g_Sprites[i];
	}
	g_Sprites.clear();
	g_Sprites.push_back(&player);
}

void SPRITE_Place(int x, int y, bool circ)
{
	Sprite *spr = new Sprite;
	spr->x = (float)x;
	spr->y = (float)y;
	spr->size = 0.6;
	spr->type = circ?0.0:1.0;
	spr->desired_type = spr->type;
	spr->health = 1.0;
	g_Sprites.push_back(spr);
}